#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import click
import pkg_resources


@click.group()
def main():
    pass


plugins = {
    entry_point.name: entry_point
    for entry_point
    in pkg_resources.iter_entry_points('neptune.plugins')
}

for name, entry_point in plugins.items():
    main.add_command(entry_point.load(), name)
